import * as kurento from 'kurento-client';
import { ConnectedSocket, MessageBody, SubscribeMessage, WebSocketGateway, WebSocketServer } from '@nestjs/websockets';
import { Server, Socket } from 'socket.io';
import { Logger } from '@nestjs/common';

@WebSocketGateway()
export class EventsGateway {
    @WebSocketServer()
    server: Server;

    private kurentoClient = null;
    private sessions = {};
    private candidatesQueue = {};

    @SubscribeMessage('start')
    handleStart(
        @MessageBody() data: string,
        @ConnectedSocket() client: Socket,
    ): void {
        const message = JSON.parse(data);
        this.start('helloWorld', client, message.sdpOffer, (error, sdpAnswer)=> {
            Logger.log('Starting Response:::');
            Logger.log(error, 'ERROR');
            if(error) {
                return client.send(JSON.stringify({
                    id: 'error',
                    message: error,
                }));
            }
            client.emit('startResponse', JSON.stringify({
                sdpAnswer: sdpAnswer,
            }));
        });
    }

    @SubscribeMessage('stop')
    handleStop(
    ): void {
        this.stop('helloWorld');
    }

    @SubscribeMessage('onIceCandidate')
    handleOnIceCandidate(@MessageBody() data: string): void{
        const message = JSON.parse(data);
        this.onIceCandidate('helloWorld', message.candidate);
    }

    private getKurentoClient(callback) {
        if(this.kurentoClient !== null) {
            return callback(null, this.kurentoClient);
        }

        kurento('ws://cliffbreak.de:8888/kurento', (error, _kurentoClient) => {
            if(error) {
                Logger.error('Could not find media server at address ' + 'ws://cliffbreak.de:8888/kurento', 'EventsGateway');
                return callback('Could not find media server at address' + 'ws://cliffbreak.de:8888/kurento'
                        + '. Exiting with error ' + error);
            }

            this.kurentoClient = _kurentoClient;
            callback(null, this.kurentoClient);
        });
    }

    private start(sessionId, ws, sdpOffer, callback) {
        if(!sessionId) {
            return callback('Cannot use undefined sessionId');
        }

        this.getKurentoClient((error, kurentoClient) => {
            if(error) {
                return callback(error);
            }

            kurentoClient.create('MediaPipeline', (error, pipeline) => {
                if(error) {
                    return callback(error);
                }
                Logger.log(error, 'Error');
                Logger.log(pipeline, 'Pipeline');

                this.createMediaElements(pipeline, ws, (error, webRtcEndpoint) => {
                    if(error) {
                        pipeline.release();
                        return callback(error);
                    }

                    if(this.candidatesQueue[sessionId]) {
                        while(this.candidatesQueue[sessionId].length) {
                            const candidate = this.candidatesQueue[sessionId].shift();
                            webRtcEndpoint.addIceCandidate(candidate);
                        }
                    }

                    this.connectMediaElements(webRtcEndpoint, (error) =>{
                        if(error) {
                            pipeline.release();
                            return callback(error);
                        }

                        webRtcEndpoint.on('OnIceCandidate', (event) =>{
                            const candidate = kurento.getComplexType('IceCandidate')(event.candidate);
                            ws.send(JSON.stringify({
                                id: 'iceCandidate',
                                candidate: candidate,
                            }));
                        });

                        webRtcEndpoint.processOffer(sdpOffer, (error, sdpAnswer) => {
                            if(error) {
                                pipeline.release();
                                return callback(error);
                            }

                            this.sessions[sessionId] = {
                                'pipeline': pipeline,
                                'webRtcEndpoint': webRtcEndpoint,
                            };
                            return callback(null, sdpAnswer);
                        });

                        webRtcEndpoint.gatherCandidates((error) => {
                            if(error) {
                                return callback(error);
                            }
                        });
                    });
                });
            });
        });
    }

    private createMediaElements(pipeline, ws, callback) {
        pipeline.create('WebRtcEndpoint', (error, webRtcEndpoint) => {
            if(error) {
                return callback(error);
            }

            return callback(null, webRtcEndpoint);
        });
    }

    private connectMediaElements(webRtcEndpoint, callback) {
        webRtcEndpoint.connect(webRtcEndpoint, (error) => {
            if(error) {
                return callback(error);
            }
            return callback(null);
        });
    }

    private stop(sessionId) {
        if(this.sessions[sessionId]) {
            const pipeline = this.sessions[sessionId].pipeline;
            Logger.log('Releasing pipeline', 'EventsGateway');
            pipeline.release();

            delete this.sessions[sessionId];
            delete this.candidatesQueue[sessionId];
        }
    }

    private onIceCandidate(sessionId, _candidate) {
        const candidate = kurento.getComplexType('IceCandidate')(_candidate);

        if(this.sessions[sessionId]) {
            Logger.log('Sending candidate', 'EventsGateway:onICECandidate');
            const webRtcEndpoint = this.sessions[sessionId].webRtcEndpoint;
            webRtcEndpoint.addIceCandidate(candidate);
        } else {
            Logger.log('Queueing candidate', 'EventsGateway:onICECandidate');
            if(!this.candidatesQueue[sessionId]) {
                this.candidatesQueue[sessionId] = [];
            }
            this.candidatesQueue[sessionId].push(candidate);
        }
    }
}